/* $Workfile: ThorAccess.c $       $Revision: 4 $      $Modtime: 5/15/12 1:02p $  */
/*******************************************************************************

  Odin TeleSystems Inc.

  OTX Telecom Adapter Driver

  @doc PublicDemo

  @module

    THORACCESS.C

    An example console application demonstrating how to access an OTX
    card even if another application has opened a handle to it.

    It demonstrates the use of direct register reads and writes to the board,
    thus in order to properly use this programe extensive knowledge of the
    the register map for the board is necessary. 

    This program is meant to be used as utility for advanced users only.

    Press 'h' or '?' for help. 'q' to quit.

*******************************************************************************/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include <assert.h>
#include <time.h>

#include "OsaStd.h"
#include "OsaCon.h"
#include "OsaThrd.h"
#include "OtxGeneric.h"
#include "OtxErr.h"
#include "OtxBrd.h"
#include "OtxDef.h"
#include "OtxType.h"
#include "OtxDrv.h"
#include "OtxHdlc.h"
#include "OtxT1E1.h"
#include "OtxData.h"
#include "OtxDsp.h"

#define DEMO_DEFAULT_LOG_FILE "ThorAccess.log"
#define DEMO_DEFAULT_REG_FILE "ThorAccess.reg"
#define DEMO_DEFAULT_BOARD_NO 0


typedef struct DemoEnvT {
  char                  *pszInputStr;
  OTX_BOOL               bExitOnError;
  char                   cwd[OTX_MAX_PATH];
  char                   logFileName[OTX_MAX_PATH];
  char                   regFileName[OTX_MAX_PATH];
  OTX_INT32              nBoardType;
  int                    boardNo;
} DemoEnvT;


/*******************************************************************************
 Function Prototypes
*******************************************************************************/
static void DemoOpenPhysicalDevices(void);
static void DemoParseCmdLineArgs(int argc, char *argv[], DemoEnvT *envP);
static void DemoHelpPanel(void);
static void DemoLogf(char *formatStr, ...);
static void DemoCheckRc(IN OTX_HANDLE hDevice, IN OTX_RESULT otxRc);
static void DemoExitQuestion(void);

static OTX_INT32 DemoProcessUserCommand(char commandChar);


/*******************************************************************************
 Global Data
*******************************************************************************/
OTX_EVENT   g_hNotificationEvent = OTX_INVALID_EVENT_VALUE;
OTX_HANDLE  g_hEventQueue  = 0;
FILE       *g_pDemoLogFile = NULL;
DemoEnvT    g_demoEnv;
OTX_BOOL    g_bHdlcVerbose = OTX_TRUE;
OTX_UINT32  g_coHdlcRxFrames[32];
OTX_UINT32  g_coHdlcBadRxFrames[32];
OTX_UINT8   g_sqHdlcFrame[OD_HDLC_MAX_TX_HDLC_FRAME_LEN];
OTX_INT32   g_nSleepTimeBetweenHdlcFrames = 0;
OTX_INT32   g_cbHdlcFrame = 10;
OTX_INT32   g_coHdlcSenderFrames = 1;

struct PhysDevs {
  OTX_HANDLE hThorController0;
  OTX_HANDLE hPlx;
} g_hPhysDevs;



/*******************************************************************************

  @func
    Open the Physical Devices on the OTX board

  @comm
    Calls function <f OtxDrvCreatePhysicalDevice> returning a handle to a
    Device Object representing a physical device, which can be the
    entire board or an integrated circuit (chip) on the board. The number of
    physical devices is set per board and cannot change dynamically.

*******************************************************************************/
void DemoOpenPhysicalDevices(void)
{
  OTX_RESULT result;
  OTX_BOOL bOpenPlxDevice = OTX_TRUE;

  // Connect to the Physical Board Controller
  switch (g_demoEnv.nBoardType) {
  case OTX_DEVICE_THOR_PCMCIA:
    DemoLogf("Opening Thor-2-PCMCIA board\n");
    break;
  case OTX_DEVICE_THOR_PCI:
    DemoLogf("Opening Thor-2-PCI board\n");
    break;
  case OTX_DEVICE_HERMOD_PCI:
    DemoLogf("Opening Hermod-PCI board\n");
    break;
  case OTX_DEVICE_THOR_1_PCI_BASIC:
    DemoLogf("Opening Thor-1-PCI-Basic board\n");
    break;
  case OTX_DEVICE_THOR_2_PCI_PLUS:
    DemoLogf("Opening Thor-2-PCI-Plus board\n");
    break;
  case OTX_DEVICE_THOR_2_PCI_EXPRESS:
    DemoLogf("Opening Thor-2-PCI-Express board\n");
    break;
  case OTX_DEVICE_THOR_1_PCI_PLUS:
    DemoLogf("Opening Thor-1-PCI-Plus board\n");
    break;
  case OTX_DEVICE_THOR_8_PCI:
    DemoLogf("Opening Thor-8-PCI board\n");
    break;
  case OTX_DEVICE_THOR_8_PCI_PLUS:
    DemoLogf("Opening Thor-8-PCI-Plus board\n");
    break;
  case OTX_DEVICE_THOR_8_PCIE:
    DemoLogf("Opening Thor-8-PCIe board\n");
    break;
  case OTX_DEVICE_THOR_4_PCI_PLUS:
    DemoLogf("Opening Thor-4-PCI-Plus board\n");
    break;
  case OTX_DEVICE_SLEIPNIR_1_PCI:
    DemoLogf("Opening Sleipnir board\n");
    break;
  case OTX_DEVICE_GIMLE_16_PCI_PLUS:
    DemoLogf("Opening Gimle-16-PCI-Plus board\n");
    break;
  case OTX_DEVICE_GIMLE_16_PCIE:
    DemoLogf("Opening Gimle-16-PCIe board\n");
    break;
  case OTX_DEVICE_GIMLE_16_PCI_BASIC_32:
    DemoLogf("Opening Gimle-16-PCI-Basic-32 board\n");
    break;
  case OTX_DEVICE_GIMLE_16_PCI_BASIC_64:
    DemoLogf("Opening Gimle-16-PCI-Basic-64 board\n");
    break;
  case OTX_DEVICE_GIMLE_16_PCIE_BASIC_32:
    DemoLogf("Opening Gimle-16-PCIe-Basic-32 board\n");
    break;
  case OTX_DEVICE_GIMLE_16_PCIE_BASIC_64:
    DemoLogf("Opening Gimle-16-PCIe-Basic-64 board\n");
    break;
  case OTX_DEVICE_HUGIN_2_PC104:
    DemoLogf("Opening Hugin-2-PC104 board\n");
    break;
  case OTX_DEVICE_ARNI_6X6_PCI_PLUS:
    DemoLogf("Opening Arni-6x6-PCI-Plus board\n");
    break;
  case OTX_DEVICE_ARNI_6X6_PCI:
    DemoLogf("Opening Arni-6x6-PCI board\n");
    break;
  case OTX_DEVICE_THOR_2_EXPRESSCARD:
    DemoLogf("Opening Thor-2-ExpressCard board\n");
    break;
  case OTX_DEVICE_THOR_4_EXPRESSCARD:
    DemoLogf("Opening Thor-4-ExpressCard board\n");
    break;
  case OTX_DEVICE_ALVIS_PCI_EXPRESS:
    DemoLogf("Opening Alvis-PCI-Express board\n");
    break;
  case OTX_DEVICE_DMP_ALVIS_0_PBX:
    DemoLogf("Opening Alvis-0-PBX board\n");
    bOpenPlxDevice = OTX_FALSE;
    break;
  case OTX_DEVICE_DMP_ALVIS_2_PBX:
    DemoLogf("Opening Alvis-2-PBX board\n");
    bOpenPlxDevice = OTX_FALSE;
    break;
  case OTX_DEVICE_DMP_ALVIS_4_PBX:
    DemoLogf("Opening Alvis-4-PBX board\n");
    bOpenPlxDevice = OTX_FALSE;
    break;
  case OTX_DEVICE_DMP_ALVIS_8_PBX:
    DemoLogf("Opening Alvis-8-PBX board\n");
    bOpenPlxDevice = OTX_FALSE;
    break;
  case OTX_DEVICE_DMP_ALVIS_0_CSI:
    DemoLogf("Opening Alvis-0-CSI board\n");
    bOpenPlxDevice = OTX_FALSE;
    break;
  case OTX_DEVICE_DMP_ALVIS_2_CSI:
    DemoLogf("Opening Alvis-2-CSI board\n");
    bOpenPlxDevice = OTX_FALSE;
    break;
  case OTX_DEVICE_DMP_ALVIS_4_CSI:
    DemoLogf("Opening Alvis-4-CSI board\n");
    bOpenPlxDevice = OTX_FALSE;
    break;
  case OTX_DEVICE_DMP_ALVIS_4M_CSI:
    DemoLogf("Opening Alvis-4M-CSI board\n");
    bOpenPlxDevice = OTX_FALSE;
    break;
  case OTX_DEVICE_DMP_ALVIS_8_CSI:
    DemoLogf("Opening Alvis-8-CSI board\n");
    bOpenPlxDevice = OTX_FALSE;
    break;
  case OTX_DEVICE_DMP_ALVIS_8M_CSI:
    DemoLogf("Opening Alvis-8M-CSI board\n");
    bOpenPlxDevice = OTX_FALSE;
    break;
  case OTX_DEVICE_DMP_ALVIS:
    DemoLogf("Opening DMP Alvis (hosted DMP)\n");
    bOpenPlxDevice = OTX_FALSE;
    break;
  case OTX_DEVICE_THOR_8M_USB:
    DemoLogf("Opening Thor-8M-USB board\n");
    bOpenPlxDevice = OTX_FALSE;
    break;
  case OTX_DEVICE_THOR_8_USB:
    DemoLogf("Opening Thor-8-USB board\n");
    bOpenPlxDevice = OTX_FALSE;
    break;

  default:
    DemoLogf("Board Type (%d) not supported by this demo program\n", g_demoEnv.nBoardType);
    OsaConsoleExit();
    exit(1);
    break;
  }

  result = OtxDrvGetDeviceHandle(0, (OtxPhysicalDeviceTypeE) g_demoEnv.nBoardType, g_demoEnv.boardNo, &g_hPhysDevs.hThorController0); 
  DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);

  DemoLogf("Board Device Handle = 0x%.8X\n", g_hPhysDevs.hThorController0);

  if (bOpenPlxDevice) {
    result = OtxDrvGetDeviceHandle(g_hPhysDevs.hThorController0, (OtxPhysicalDeviceTypeE) OTX_DEVICE_PCI_BRIDGE, 0, &g_hPhysDevs.hPlx); 

    DemoLogf("Plx Device Handle = 0x%.8X\n", g_hPhysDevs.hPlx);
  }

}



/*******************************************************************************

  @func

  Parse the program's command line arguments and calls upon
  requested actions.

  There are some optional switches that can be used following
  the driver name on the command line. Switches are are
  separated by spaces, and can appear in arbitrary order.
  Switches consist of a hyphen and a letter:

  @comm

*******************************************************************************/
static void DemoParseCmdLineArgs(int argc, char *argv[], DemoEnvT *envP) 
{
  short  i;

  // Set reasonable defaults before reading from command line
  envP->pszInputStr       = NULL;
  envP->bExitOnError      = OTX_FALSE;
  envP->nBoardType        = OTX_DEVICE_UNKNOWN;
  envP->boardNo           = DEMO_DEFAULT_BOARD_NO;

  strcpy(envP->logFileName, envP->cwd);
  strcat(envP->logFileName, DEMO_DEFAULT_LOG_FILE);
  strcat(envP->regFileName, DEMO_DEFAULT_REG_FILE);
  for (i = 1; i < argc; i++) {        // Scan command line arguments
    if (*(argv[i]+1) == 'Z') {   // Command input string -- case sensitive
      envP->pszInputStr = argv[i]+2;
    } else {    
      OsaStdStrUpr(argv[i]);
      if (*argv[i] == '-' || *argv[i] == '/') {  // Is argument a legal switch?
        if (*(argv[i]+1) == 'L') {   // Log file
          strcpy(envP->logFileName, argv[i] + 2);
        } else if (*(argv[i]+1) == 'R') {   // Register/Value file
          strcpy(envP->regFileName, argv[i] + 2);
        } else if (*(argv[i]+1) == 'Q') {   // Exit on error
          envP->bExitOnError = OTX_TRUE;
        } else if (*(argv[i]+1) == 'B') {   // Board number
          envP->boardNo = strtoul(argv[i]+2, NULL, 0);
        } else if (*(argv[i]+1) == 'T') {   // Board Type
          if ( *(argv[i]+2) == '2') {
            envP->nBoardType = OTX_DEVICE_THOR_PCI;
          } else if ( *(argv[i]+2) == '9') {
            envP->nBoardType = OTX_DEVICE_HERMOD_PCI;
          } else if (*(argv[i] + 2) == '8') {
            envP->nBoardType  = OTX_DEVICE_THOR_8_PCI;
          } else if (*(argv[i] + 2) == 'T') {
            envP->nBoardType  = OTX_DEVICE_THOR_2_PCI_PLUS;
          } else if (*(argv[i] + 2) == 'E') {
            envP->nBoardType  = OTX_DEVICE_THOR_2_PCI_EXPRESS;
          } else if (*(argv[i] + 2) == 'P') {
            envP->nBoardType  = OTX_DEVICE_THOR_8_PCI_PLUS;
          } else if (*(argv[i] + 2) == 'W') {
            envP->nBoardType  = OTX_DEVICE_THOR_8_PCIE;
          } else if (*(argv[i] + 2) == '4') {
            envP->nBoardType  = OTX_DEVICE_THOR_4_PCI_PLUS;
          } else if (*(argv[i] + 2) == 'M') {
            envP->nBoardType  = OTX_DEVICE_THOR_PCMCIA;
          } else if (*(argv[i] + 2) == 'S') {
            envP->nBoardType  = OTX_DEVICE_SLEIPNIR_1_PCI;
          } else if (*(argv[i] + 2) == 'B') {
            envP->nBoardType  = OTX_DEVICE_THOR_1_PCI_BASIC;
          } else if (*(argv[i] + 2) == '1') {
            envP->nBoardType  = OTX_DEVICE_THOR_1_PCI_PLUS;
          } else if (*(argv[i] + 2) == 'G') {
            envP->nBoardType  = OTX_DEVICE_GIMLE_16_PCI_PLUS;
          } else if (*(argv[i] + 2) == '9') {
            envP->nBoardType  = OTX_DEVICE_GIMLE_16_PCIE;
          } else if (*(argv[i] + 2) == '3') {
            envP->nBoardType  = OTX_DEVICE_GIMLE_16_PCI_BASIC_32;
          } else if (*(argv[i] + 2) == '6') {
            envP->nBoardType  = OTX_DEVICE_GIMLE_16_PCI_BASIC_64;
          } else if (*(argv[i] + 2) == '5') {
            envP->nBoardType  = OTX_DEVICE_GIMLE_16_PCIE_BASIC_32;
          } else if (*(argv[i] + 2) == '7') {
            envP->nBoardType  = OTX_DEVICE_GIMLE_16_PCIE_BASIC_64;
          } else if (*(argv[i] + 2) == 'H') {
            envP->nBoardType  = OTX_DEVICE_HUGIN_2_PC104;
          } else if (*(argv[i] + 2) == 'A') {
            envP->nBoardType  = OTX_DEVICE_ARNI_6X6_PCI_PLUS;
          } else if (*(argv[i] + 2) == 'J') {
            envP->nBoardType  = OTX_DEVICE_ARNI_6X6_PCI;
          } else if (*(argv[i] + 2) == 'C') {
            envP->nBoardType  = OTX_DEVICE_THOR_2_EXPRESSCARD;
          } else if (*(argv[i] + 2) == 'R') {
            envP->nBoardType  = OTX_DEVICE_THOR_4_EXPRESSCARD;
          } else if (*(argv[i] + 2) == 'D') {
            envP->nBoardType  = OTX_DEVICE_ALVIS_PCI_EXPRESS;          
          } else if (*(argv[i] + 2) == 'I') {
            envP->nBoardType  = OTX_DEVICE_DMP_ALVIS_4_CSI;          
          } else if (*(argv[i] + 2) == 'K') {
            envP->nBoardType  = OTX_DEVICE_DMP_ALVIS_4M_CSI;          
          } else if (*(argv[i] + 2) == 'X') {
            envP->nBoardType  = OTX_DEVICE_DMP_ALVIS_8_CSI;          
          } else if (*(argv[i] + 2) == 'Q') {
            envP->nBoardType  = OTX_DEVICE_DMP_ALVIS_8M_CSI;          
          } else if (*(argv[i] + 2) == 'V') {
            envP->nBoardType  = OTX_DEVICE_DMP_ALVIS;          
          } else if (*(argv[i] + 2) == '0') {
            envP->nBoardType  = OTX_DEVICE_DMP_ALVIS_0_CSI;          
          } else if (*(argv[i] + 2) == 'W') {
            envP->nBoardType  = OTX_DEVICE_DMP_ALVIS_2_CSI;          
          } else if (*(argv[i] + 2) == 'Y') {
            envP->nBoardType  = OTX_DEVICE_DMP_ALVIS_0_PBX;          
          } else if (*(argv[i] + 2) == 'Z') {
            envP->nBoardType  = OTX_DEVICE_DMP_ALVIS_2_PBX;          
          } else if (*(argv[i] + 2) == 'L') {
            envP->nBoardType  = OTX_DEVICE_DMP_ALVIS_4_PBX;          
          } else if (*(argv[i] + 2) == 'N') {
            envP->nBoardType  = OTX_DEVICE_DMP_ALVIS_8_PBX;          
          } else if (*(argv[i] + 2) == 'U') {
            envP->nBoardType  = OTX_DEVICE_THOR_8M_USB;          
          } else if (*(argv[i] + 2) == '1') {
            envP->nBoardType  = OTX_DEVICE_THOR_8_USB;          
          } else {
            DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
            DemoHelpPanel();
            OsaConsoleExit();
            exit(1);
          }
        } else {
          DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
          DemoHelpPanel();
          OsaConsoleExit();
          exit(1);
        }
      } else {
        DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
        DemoHelpPanel();
        OsaConsoleExit();
			  exit(1);
      }
    }
  }
}


/*******************************************************************************

  @func

  Displays a help for the command line.

  @comm

*******************************************************************************/
static void DemoHelpPanel()
{
  DemoLogf("\nCommand line switches:\n");
  DemoLogf(" /B<boardNo>    : Board number (Default is %d)\n", DEMO_DEFAULT_BOARD_NO);
  DemoLogf(" /T[B|2|T|E|8|P|W|4|M|S|G|9|3|6|5|7|H|A|J|C|R|D|I|K|X|Q|V|0|W|Y|Z|L|N|U] : Board Type used for the demo, where\n");
  DemoLogf("                    B  = Use Thor-1-PCI-Basic Board\n");
  DemoLogf("                    2  = Use Thor-2-PCI Board\n");
  DemoLogf("                    9  = Use Hermod-PCI Board\n");
  DemoLogf("                    T  = Use Thor-2-PCI-Plus Board\n");
  DemoLogf("                    E  = Use Thor-2-PCI-Express Board\n");
  DemoLogf("                    8  = Use Thor-8-PCI Board\n");
  DemoLogf("                    P  = Use Thor-8-PCI-Plus Board\n");
  DemoLogf("                    W  = Use Thor-8-PCIe Board\n");
  DemoLogf("                    4  = Use Thor-4-PCI-Plus Board\n");
  DemoLogf("                    M  = Use Thor-2-PCMCIA Board\n");
  DemoLogf("                    S  = Use Sleipnir Board\n");
  DemoLogf("                    G  = Use Gimle-16-PCI-Plus Board\n");
  DemoLogf("                    9  = Use Gimle-16-PCIe Board\n");
  DemoLogf("                    3  = Use Gimle-16-PCI-Basic_32 Board\n");
  DemoLogf("                    6  = Use Gimle-16-PCI-Basic_64 Board\n");
  DemoLogf("                    5  = Use Gimle-16-PCIe-Basic_32 Board\n");
  DemoLogf("                    7  = Use Gimle-16-PCIe-Basic_64 Board\n");
  DemoLogf("                    H  = Use Hugin-2-PC104 Board\n");
  DemoLogf("                    A  = Use Arni-6-PCI-Plus Board\n");
  DemoLogf("                    J  = Use Arni-6-PCI Board\n");
  DemoLogf("                    C  = Use Thor-2-ExpressCard Board\n");
  DemoLogf("                    R  = Use Thor-4-ExpressCard Board\n");
  DemoLogf("                    D  = Use Alvis-PCI-Express Board\n");
  DemoLogf("                    I  = Use Alvis-4-CSI Board\n");
  DemoLogf("                    K  = Use Alvis-4M-CSI Board\n");
  DemoLogf("                    X  = Use Alvis-8-CSI Board\n");
  DemoLogf("                    Q  = Use Alvis-8M-CSI Board\n");
  DemoLogf("                    V  = Use Alvis-DMP (hosted DMP)\n");
  DemoLogf("                    0  = Use Alvis-0-CSI Board\n");
  DemoLogf("                    W  = Use Alvis-2-CSI Board\n");
  DemoLogf("                    Y  = Use Alvis-0-PBX Board\n");
  DemoLogf("                    Z  = Use Alvis-2-PBX Board\n");
  DemoLogf("                    L  = Use Alvis-4-PBX Board\n");
  DemoLogf("                    N  = Use Alvis-8-PBX Board\n");
  DemoLogf("                    U  = Use Thor-8M-USB Board\n");
  DemoLogf("                    1  = Use Thor-8-USB Board\n");
  DemoLogf(" /L<fileName>   : Log file name\n");
  DemoLogf(" /R<fileName>   : Register/Value file name\n");
  DemoLogf(" /Z<str>        : Command string (overrides keyboard input)\n");
  DemoLogf(" /Q             : Exit without prompting on errors\n");

  DemoLogf("\nPress Any Key...\n"); OsaConsoleGetCh();

}


/*******************************************************************************

  @func

  Displays a message on the screen and prints it to the logfile.

  @comm

*******************************************************************************/
static void DemoLogf(char *formatStr, ...)
{
  va_list argList;
  va_start(argList, formatStr);

  OsaConsolePrintF(formatStr, argList);
  if (g_pDemoLogFile) vfprintf(g_pDemoLogFile,   formatStr, argList);

  va_end(argList);

}


/*******************************************************************************

  @func

  Displays a message on the screen 

  @comm

*******************************************************************************/
static void DemoPrintf(char *formatStr, ...)
{
  va_list argList;
  va_start(argList, formatStr);

  OsaConsolePrintF(formatStr, argList);

  va_end(argList);

}

static void DemoCheckRc(IN OTX_HANDLE hDevice, IN OTX_RESULT result)
/*******************************************************************************
  @func
    Check the result from API call. If the result is a failure code then prompt
    the user if he wants to exit.
*******************************************************************************/
{
  if ( OTX_FAILURE(result) ) {

    if (hDevice == OTX_INVALID_HANDLE_VALUE) {
      DemoLogf("ERROR(%#lX [%d.%d]) \n", 
               OTX_RESULT_CODE(result), 
               OTX_RESULT_SOURCE_ID(result), 
               OTX_RESULT_LINE_NUM(result) );
    } else {
      OTX_RESULT otxRc;
      char msgStr[256];
      OtxAttribTypeE  eAttribType;    // Type of the attribute, if attribute exists
      OTX_UINT32      cbAttribValue;  // Size in bytes of attribute value (Get value with <f OtxDrvGetAttributeValue>)
      OTX_BOOL        bReadOnly;      // Is attribute writable or read-only
      OTX_UINT32      eDeviceType = 999;    // Device type for hDevice 
      OTX_UINT32      nDeviceNumber = 999;  // Device number for hDevice
      
      // Get the Device Type for hHandle
      otxRc = OtxDrvGetAttributeType(hDevice, OTX_ATTR_DEVICE_TYPE, &eAttribType, &cbAttribValue, &bReadOnly); 
      if (OTX_SUCCESS(otxRc)) {
        assert(eAttribType == OTX_ATTRIB_UINT32);
        assert(cbAttribValue == sizeof(OTX_UINT32));
        OtxDrvGetAttributeValueUint32(hDevice, OTX_ATTR_DEVICE_TYPE, &eDeviceType); 
      } 
      
      // Get the Device Number for hHandle
      otxRc = OtxDrvGetAttributeType(hDevice, OTX_ATTR_DEVICE_NO, &eAttribType, &cbAttribValue, &bReadOnly); 
      if (OTX_SUCCESS(otxRc)) {
        assert(eAttribType == OTX_ATTRIB_UINT32);
        assert(cbAttribValue == sizeof(OTX_UINT32));
        OtxDrvGetAttributeValueUint32(hDevice, OTX_ATTR_DEVICE_NO, &nDeviceNumber); 
      }
      
      // Convert the result code to a more descriptive string
      otxRc = OtxDrvResultCode2String(OTX_DRIVER_DEVICE_HANDLE, result, 256, &msgStr[0]);
      if (OTX_FAILURE(otxRc)) {
        strcpy(msgStr, "Cannot translate this result code.");
      }
      DemoLogf("ERROR(%#lX [%d.%d]) Dev(%X,%d,%d): %s \n", OTX_RESULT_CODE(result), OTX_RESULT_SOURCE_ID(result), OTX_RESULT_LINE_NUM(result), hDevice, (int) eDeviceType, (int)nDeviceNumber, msgStr);
    }

    DemoExitQuestion();
  }
}

/*******************************************************************************
  @func
    Prints a timestamp in a MM-DD hh:mm:ss:mmm format, where

    MM  = month
    DD  = day (of month)
    hh  = hour of day
    mm  = minutes of day
    ss  = seconds of day
    mmm = milliseconds

*******************************************************************************/
void DemoPrintTimeStamp(
  OTX_DATETIME nTimeStamp
)
{
  OTX_DATETIME nTimeStampMs;
  OTX_DATETIME nTimeStampSec;
  struct Osa_tm *ptCurrentDayAndTime;
  Osa_time_t nCurrentTimeSecs;
  
  // Subtract number of 100ns ticks between Jan 1, 1970 and Jan 1, 1600
  
  #if defined(__linux__)
    nTimeStamp.nUint64 = nTimeStamp.nUint64 - 1164447360 * 100000000;  
  #elif defined(WIN32)
    nTimeStamp.nUint64 = nTimeStamp.nUint64 - 116444736000000000;  
  #endif

  
  // Convert to seconds since Jan 1, 1970 (so we can use this time with OsaStdLocalTime())
  nTimeStampSec.nUint64 = nTimeStamp.nUint64 / 10000000;
  
  // Calculate number of milliseconds (since number of seconds is not a very accurate way to report timestamp)
  nTimeStampMs.nUint64 = nTimeStamp.nUint64 / 10000 - (nTimeStampSec.nUint64 * 1000);
  
  nCurrentTimeSecs = nTimeStampSec.sUint32.nLow;
  ptCurrentDayAndTime = OsaStdLocalTime(&nCurrentTimeSecs);
  
  if (ptCurrentDayAndTime) {
    
    DemoLogf("%.2d-%.2d %.2d:%.2d:%.2d.%.3d ", 
      ptCurrentDayAndTime->tm_mon + 1,
      ptCurrentDayAndTime->tm_mday,
      ptCurrentDayAndTime->tm_hour,
      ptCurrentDayAndTime->tm_min, 
      ptCurrentDayAndTime->tm_sec,
      nTimeStampMs.sUint32.nLow
      );
  } else {
    DemoLogf("Error: Invalid timestamp (before Jan 1, 1970)\n");
  }
  
  
}

/*******************************************************************************
  @func
    Prompts whether to exit the program. Usually called after an error 
    has been reported.

*******************************************************************************/
void DemoExitQuestion()
{
  short c;

  if (g_demoEnv.bExitOnError) {
    c = 'y';
  } else {
    DemoLogf("Do you want to exit (Y/N) ");
    c = OsaConsoleGetCh();
  }
  if (c == 'y' || c == 'Y') {
    // Clean up
    OtxDrvReleaseLib();
    OsaConsoleExit();
    exit(1);
  }
}


/*******************************************************************************

  @func

  Processes User Commands

  @comm

*******************************************************************************/
OTX_INT32 DemoProcessUserCommand(char commandChar)
{
  OTX_RESULT result;



  switch (commandChar) {

  case '?':
  case 'h':
  case 'H':
    DemoLogf("Cmd  Synopsis\n");
    DemoLogf("===  ========\n");
    DemoLogf(" ?   Display Help text\n");
    DemoLogf(" h   Ditto\n");
    DemoLogf(" q   Quit\n");
    DemoLogf(" r   OtxDrvRead8 from board\n");
    DemoLogf(" w   OtxDrvWriteReg8 to board\n");
    DemoLogf(" i   OtxDrvRead16 from board\n");
    DemoLogf(" o   OtxDrvWriteReg16 to board\n");
    DemoLogf(" R   OtxDrvRead32 from board\n");
    DemoLogf(" W   OtxDrvWriteReg32 to board\n");
    DemoLogf(" p   OtxDrvRead32 from Pci bridge device\n");
    DemoLogf(" d   Issue Debug Command OtxDrvDbgDump() to board\n");
    DemoLogf(" t   Print current time\n");
    DemoLogf(" T   Prompt for timestamp in hex. Print it in readable format.\n");
    DemoLogf(" M   Memory Read/Write Test.\n");
    DemoLogf(" B   Bulk OtxDrvWriteReg8 from file.\n");
    break;

  case 'B':
    {
      FILE *fin;
      char strBuf[128];
      OTX_UINT32 nAddress;
      OTX_UINT32 nValue;
      char *token;

      if ((fin = fopen(g_demoEnv.regFileName, "rt")) == NULL) {
        DemoLogf("Couldn't open regfile %s\n", g_demoEnv.regFileName);
        break;
      }

      while (fgets(strBuf, 128, fin)) {
        token = strtok(strBuf, " ,\t\n");
        if (!token) {
          DemoLogf("Unexpected Null token. Operation aborted.\n");
          fclose(fin);
          break;
        }
        nAddress = strtoul(token, NULL, 0);
        token = strtok(NULL, " ,\t\n");
        if (!token) {
          DemoLogf("Unexpected Null token [2]. Operation aborted.\n");
          fclose(fin);
          break;
        }
        nValue = strtoul(token, NULL, 0);

        result = OtxDrvWriteReg8(g_hPhysDevs.hThorController0, nAddress, nValue); DemoCheckRc(g_hPhysDevs.hThorController0, result);

        DemoLogf("Wrote %.2X to Brd Address 0x%X\n", (OTX_UINT8)nValue, nAddress);
      }
      fclose(fin);

    }
    break;


  case 't':
    {
      OTX_DATETIME  nTimeStamp;

      result = OtxDrvGetCurrentTime(&nTimeStamp);

      DemoPrintTimeStamp(nTimeStamp);
      DemoLogf("\n");

    }
    break;

  case 'T':
    {
      OTX_DATETIME  nTimeStamp;
      char strBuf[64];
      static OTX_UINT32 nTimeStampHigh = 0;
      static OTX_UINT32 nTimeStampLow = 0;

      DemoLogf("Enter High 32-bit value of timestamp (0x%.8X): ", nTimeStampHigh);
      OsaConsoleGetStr(strBuf); 
      if (*strBuf) {
        nTimeStampHigh = strtoul(strBuf, NULL, 0);
      }

      DemoLogf("Enter Low 32-bit value of timestamp (0x%.8X): ", nTimeStampLow);
      OsaConsoleGetStr(strBuf); 
      if (*strBuf) {
        nTimeStampLow = strtoul(strBuf, NULL, 0);
      }

      nTimeStamp.sUint32.nHigh = nTimeStampHigh;
      nTimeStamp.sUint32.nLow  = nTimeStampLow;

      DemoPrintTimeStamp(nTimeStamp);
      DemoLogf("\n");

    }
    break;

  case 'r':
    {
      OTX_UINT32 readVal;
      char strBuf[12];
      static OTX_UINT32 nAddress = 0;

      DemoLogf("Read from a 8-bit register in the OTX Board device.\n");
      DemoLogf("Which Address (0x%X): ", nAddress);
      OsaConsoleGetStr(strBuf); 
      if (*strBuf) {
        nAddress = strtoul(strBuf, NULL, 0);
      }

      result = OtxDrvReadReg8(g_hPhysDevs.hThorController0, nAddress, &readVal); DemoCheckRc(g_hPhysDevs.hThorController0, result);

      DemoLogf("Brd Address 0x%X = 0x%.2X (%d decimal)\n", nAddress, readVal, readVal);
    }
    break;

  case 'M':
    {
      char strBuf[12];
      static OTX_UINT32 nAddress = 0;
      static OTX_UINT32 nRange = 0;
      OTX_UINT32 nWriteVal;
      OTX_UINT32 nReadVal;
      OTX_UINT32 nErrors = 0;
      OTX_UINT32 i;

      DemoLogf("Memory Read/Write test.\n");
      DemoLogf("Which Start Address (0x%X): ", nAddress);
      OsaConsoleGetStr(strBuf);   
      if (*strBuf) {
        nAddress = strtoul(strBuf, NULL, 0);
      }
      DemoLogf("Which Range (in bytes) (0x%X): ", nRange);
      OsaConsoleGetStr(strBuf);   
      if (*strBuf) {
        nRange = strtoul(strBuf, NULL, 0);
      }

      for (i = 0; i < nRange; i += 4) {
        nWriteVal = nAddress + i;
        result = OtxDrvWriteReg32(g_hPhysDevs.hThorController0, nAddress + i, nWriteVal); DemoCheckRc(g_hPhysDevs.hThorController0, result);
      }

      for (i = 0; i < nRange; i += 4) {
        nWriteVal = nAddress + i;
        result = OtxDrvReadReg32(g_hPhysDevs.hThorController0, nAddress + i, &nReadVal); DemoCheckRc(g_hPhysDevs.hThorController0, result);

        if (nReadVal != nWriteVal) {

          DemoLogf("Memory error at Address 0x%X   (nReadVal(0x%X) != nWriteVal(0x%X))\n", nAddress + i, nReadVal, nWriteVal);
          nErrors++;
        }
      }

      DemoLogf("Memory test finished. ErrorCount=%d\n", nErrors);
    }
    break;

  case 'w':
    {
      char strBuf[12];
      static OTX_UINT32 nAddress = 0;
      static OTX_UINT32 nValue = 0;

      DemoLogf("Write to a 8-bit register in the OTX board device.\n");
      DemoLogf("Which Address (0x%X): ", nAddress);
      OsaConsoleGetStr(strBuf);   
      if (*strBuf) {
        nAddress = strtoul(strBuf, NULL, 0);
      }
      DemoLogf("Which Value (0x%X): ", nValue);
      OsaConsoleGetStr(strBuf);   
      if (*strBuf) {
        nValue = strtoul(strBuf, NULL, 0);
      }

      result = OtxDrvWriteReg8(g_hPhysDevs.hThorController0, nAddress, nValue); DemoCheckRc(g_hPhysDevs.hThorController0, result);

      DemoLogf("Wrote %.2X to Brd Address 0x%X\n", (OTX_UINT8)nValue, nAddress);
    }
    break;

  case 'i':
    {
      OTX_UINT32 readVal;
      char strBuf[12];
      static OTX_UINT32 nAddress = 0;

      DemoLogf("Read from a 16-bit register in the OTX Board device.\n");
      DemoLogf("Which Address (0x%X): ", nAddress);
      OsaConsoleGetStr(strBuf); 
      if (*strBuf) {
        nAddress = strtoul(strBuf, NULL, 0);
      }

      result = OtxDrvReadReg16(g_hPhysDevs.hThorController0, nAddress, &readVal); DemoCheckRc(g_hPhysDevs.hThorController0, result);

      DemoLogf("Brd Address 0x%X = 0x%.4X (%d decimal)\n", nAddress, readVal, readVal);
    }
    break;

  case 'o':
    {
      char strBuf[12];
      static OTX_UINT32 nAddress = 0;
      static OTX_UINT32 nValue = 0;

      DemoLogf("Write to a 16-bit register in the OTX board device.\n");
      DemoLogf("Which Address (0x%X): ", nAddress);
      OsaConsoleGetStr(strBuf);   
      if (*strBuf) {
        nAddress = strtoul(strBuf, NULL, 0);
      }
      DemoLogf("Which Value (0x%X): ", nValue);
      OsaConsoleGetStr(strBuf);   
      if (*strBuf) {
        nValue = strtoul(strBuf, NULL, 0);
      }

      result = OtxDrvWriteReg16(g_hPhysDevs.hThorController0, nAddress, nValue); DemoCheckRc(g_hPhysDevs.hThorController0, result);

      DemoLogf("Wrote %.4X to Brd Address 0x%X\n", (OTX_UINT16)nValue, nAddress);
    }
    break;

  case 'R':
    {
      OTX_UINT32 readVal;
      char strBuf[12];
      static OTX_UINT32 nAddress = 0;

      DemoLogf("Read from a 32-bit register in the OTX Board device.\n");
      DemoLogf("Which Address (0x%X): ", nAddress);
      OsaConsoleGetStr(strBuf); 
      if (*strBuf) {
        nAddress = strtoul(strBuf, NULL, 0);
      }

      result = OtxDrvReadReg32(g_hPhysDevs.hThorController0, nAddress, &readVal); DemoCheckRc(g_hPhysDevs.hThorController0, result);

      DemoLogf("Brd Address 0x%X = 0x%.8X (%d decimal)\n", nAddress, readVal, readVal);
    }
    break;

  case 'W':
    {
      char strBuf[12];
      static OTX_UINT32 nAddress = 0;
      static OTX_UINT32 nValue = 0;

      DemoLogf("Write to a 32-bit register in the OTX board device.\n");
      DemoLogf("Which Address (0x%X): ", nAddress);
      OsaConsoleGetStr(strBuf);   
      if (*strBuf) {
        nAddress = strtoul(strBuf, NULL, 0);
      }
      DemoLogf("Which Value (0x%X): ", nValue);
      OsaConsoleGetStr(strBuf);   
      if (*strBuf) {
        nValue = strtoul(strBuf, NULL, 0);
      }

      result = OtxDrvWriteReg32(g_hPhysDevs.hThorController0, nAddress, nValue); DemoCheckRc(g_hPhysDevs.hThorController0, result);

      DemoLogf("Wrote %.8X to Brd Address 0x%X\n", (OTX_UINT32)nValue, nAddress);
    }
    break;

  case 'p':
    {
      OTX_UINT32 readVal;
      char strBuf[12];
      static OTX_UINT32 nAddress = 0;

      if (g_hPhysDevs.hPlx) {
        
        DemoLogf("Read from a 32-bit register in the PCI Bridge device.\n");
        DemoLogf("Which Address (0x%X): ", nAddress);
        OsaConsoleGetStr(strBuf); 
        if (*strBuf) {
          nAddress = strtoul(strBuf, NULL, 0);
        }
        
        result = OtxDrvReadReg32(g_hPhysDevs.hPlx, nAddress, &readVal); DemoCheckRc(g_hPhysDevs.hPlx, result);
        
        DemoLogf("Plx Address 0x%X = 0x%.8X (%d decimal)\n", nAddress, readVal, readVal);
      } else {
        DemoLogf("No Plx device on this board\n");
      }
    }
    break;

  case 'P':
    {
      OTX_UINT32 readVal;
      static OTX_UINT32 nAddress = 0;

      if (g_hPhysDevs.hPlx) {
        DemoLogf("Complete register dump from the PCI Bridge device.\n");
        
        for (nAddress = 0; nAddress < 0x200; nAddress += 4) {
          result = OtxDrvReadReg32(g_hPhysDevs.hPlx, nAddress, &readVal); DemoCheckRc(g_hPhysDevs.hPlx, result);
          DemoLogf("Plx Address 0x%X = 0x%.8X (%d decimal)\n", nAddress, readVal, readVal);
        }
      } else {
        DemoLogf("No Plx device on this board\n");
      }
    }
    break;

  case 'S':
    {
      char strBuf[12];
      static OTX_UINT32 nAddress = 0;
      static OTX_UINT32 nValue = 0;

      if (g_hPhysDevs.hPlx) {
        DemoLogf("Write to a 32-bit register in the PCI Bridge device.\n");
        DemoLogf("Which Address (0x%X): ", nAddress);
        OsaConsoleGetStr(strBuf);   
        if (*strBuf) {
          nAddress = strtoul(strBuf, NULL, 0);
        }
        DemoLogf("Which Value (0x%X): ", nValue);
        OsaConsoleGetStr(strBuf);   
        if (*strBuf) {
          nValue = strtoul(strBuf, NULL, 0);
        }
        
        result = OtxDrvWriteReg32(g_hPhysDevs.hPlx, nAddress, nValue); DemoCheckRc(g_hPhysDevs.hPlx, result);
        
        DemoLogf("Wrote %.8X to PCI Bridge Address 0x%X\n", (OTX_UINT16)nValue, nAddress);
      } else {
        DemoLogf("No Plx device on this board\n");
      }
    }
    break;

  case 'd':
    {
      char szBuf[12];
      static OTX_INT32 nCmd = 0;
      
      DemoLogf("Driver Debug Command. Which Cmd? [0x%.8X] ", nCmd);
      OsaConsoleGetStr(szBuf);
      
      if (szBuf[0]) {
        nCmd = (OTX_INT32) strtol(szBuf, NULL, 0);
      }
      
      DemoLogf("Issuing Driver Debug Command 0x%.8X\n", nCmd);

      result = OtxDrvDbgDump(g_hPhysDevs.hThorController0, nCmd);
      DemoCheckRc(g_hPhysDevs.hThorController0, result);
    }
    break;


  case '~':
    {
      char strBuf[12];
      OTX_INT32 nSleepTime;

      DemoPrintf("Enter milliseconds to sleep :");
      OsaConsoleGetStr(strBuf);

      if (strBuf[0]) {
        nSleepTime = strtol(strBuf, NULL, 0);
        DemoPrintf("Sleeping %d milliseconds\n", nSleepTime);
        OsaThreadSleep(nSleepTime);
      }
    }
    break;


  case 'q':
  case 'Q':
    return 1;
    break;

  default:
    DemoLogf("Unknown command. Press '?' for help, 'q' to quit\n");
    break;
  }

  return 0;
}


/*******************************************************************************
*******************************************************************************/
void DemoAtExit(void)
{
  fclose(g_pDemoLogFile);
}


#if 0
// Returns the string representing the device type for a specific device, automatically chooses between logical and physical devices
static const char *DemoGetDeviceTypeString(IN OTX_HANDLE hDevice)
{
  static char szBuf[256];
  OTX_RESULT otxRc;
  
  otxRc = OtxDrvGetAttributeValueSz(hDevice, OTX_ATTR_DEVICE_STRING, szBuf, 256);
  if (!OTX_SUCCESS(otxRc)) {
      DemoLogf("Could not get the OTX_ATTR_DEVICE_STRING attribute for device 0x%X\n", hDevice);
      return "*** Error ***";
  } else {
    return szBuf;
  }
}

// Returns the device no for a specific device
static signed int DemoGetDeviceNo(IN OTX_HANDLE hDevice)
{
  OTX_UINT32 nDevNo;
  if (OTX_SUCCESS(OtxDrvGetAttributeValueUint32(hDevice, OTX_ATTR_DEVICE_NO, &nDevNo))) {
    return nDevNo;
  } else {
    return -1;
  }
}
#endif


/*******************************************************************************

  @func

  Enumerates all devices for a given parent devices and prints attributes
  etc.

  @comm

*******************************************************************************/
static OTX_RESULT FindBoardDevice(OTX_INT32 *peBrdType) 
{
  OTX_UINT32   eDevType;
  OTX_ITERATOR itrChildTypes;
  OTX_ITERATOR itrChildrenOfSameType;
  static char szBuf[256];
  OTX_RESULT otxRc;

  *peBrdType = OTX_DEVICE_UNKNOWN; // Defensive
  
  // Initialize the opaque iterator for the hosted types
  OtxDrvIteratorBegin(&itrChildTypes);
  while (OTX_SUCCESS(OtxDrvEnumDeviceType(0, &itrChildTypes, (OTX_INT32 *)&eDevType))) {   // Note that we must use a zero handle to get the children of the root level
    OTX_HANDLE hChildDevice;
    // Initialize the opaque iterator for the devices within the same type
    OtxDrvIteratorBegin(&itrChildrenOfSameType);
    while (OTX_SUCCESS(OtxDrvEnumDevice(0, eDevType, &itrChildrenOfSameType, &hChildDevice))) {
      if (hChildDevice != OTX_INVALID_HANDLE_VALUE) {

				// Here we have a handle for a child device, get the attributes
#if 0
        DemoLogf("%s_%d (%s, handle = %#x)\n", 
          DemoGetDeviceTypeString(hChildDevice), 
          DemoGetDeviceNo(hChildDevice), 
          (eDevType >= 0) ? (eDevType == 0 ? "UnKnown" : "Physical") : "Logical", 
          hChildDevice);
#endif

        // Check device type
        switch (eDevType) {
          // These devices appear at the top (driver level) but they are not board devices, so ignore them here
        case OTX_DEVICE_OTPD_INTERFACE:
        case OTX_DEVICE_CT_BUS:
          break;
        default:
          otxRc = OtxDrvGetAttributeValueSz(hChildDevice, OTX_ATTR_DEVICE_STRING, szBuf, 256);
          if (!OTX_SUCCESS(otxRc)) {
            DemoLogf("Could not get the OTX_ATTR_DEVICE_STRING attribute for device 0x%X\n", hChildDevice);
            return otxRc;
          } 
          DemoLogf("Found %s board (devType=%d)\n", szBuf, eDevType);
          *peBrdType = eDevType;
          return OTX_S_OK;
        }

      }
    }
  }

  return OTX_S_FALSE; // We did not find an OTX board
}


/*******************************************************************************
*******************************************************************************/
int main(int argc, char *argv[])
{
  OTX_INT32 exitFlag = 0;
  OTX_RESULT result;  // OTX Standard Return Code
  char commandChar;

  OsaConsoleInit();

  DemoLogf("THORACCESS, %s, Odin TeleSystems Inc.\n", __DATE__);
  DemoLogf("Board Device Access Example program.\n");
  DemoParseCmdLineArgs(argc, argv, &g_demoEnv);

  if (g_demoEnv.pszInputStr) {
    OsaConsoleInput(g_demoEnv.pszInputStr);
  }

  if ((g_pDemoLogFile = fopen(g_demoEnv.logFileName, "wt")) == NULL) {
    printf("Couldn't open logfile %s\n", g_demoEnv.logFileName);
    OsaConsoleExit();
		exit(-1);
  }
  atexit(DemoAtExit);

  DemoLogf("Log file is  %s\n", g_demoEnv.logFileName);
  DemoLogf("Board number %d\n", g_demoEnv.boardNo);


  // Query Library revision
  {
    char szVersionInfo[255];
    OTX_BOOL bIsStaticLinkedLibrary;
    char szLibFileName[64];
    result = OtxDrvGetAttributeValueSz(OTX_LIB_HANDLE, OTX_ATTR_LIB_REV_STR, szVersionInfo, 255); DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
    result = OtxDrvGetAttributeValueBool(OTX_LIB_HANDLE, OTX_ATTR_LIB_TYPE_STATIC, &bIsStaticLinkedLibrary); DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
    result = OtxDrvGetAttributeValueSz(OTX_LIB_HANDLE, OTX_ATTR_LIB_FILE_NAME, szLibFileName, 64); DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
    
    DemoLogf("Library version %s (filename %s, %s linked)\n", szVersionInfo, szLibFileName, bIsStaticLinkedLibrary ? "static":"dynamic");
  }

  // Connect to the Driver
  DemoLogf("Connecting lib...\n");
  result = OtxDrvConnectLib(); 
  switch (OTX_RESULT_CODE(result)) {
  case OTX_RES_HANDLE:
    DemoLogf("Error: Could not connect to the OTX library.\nPossible cause: OTX Driver not loaded\n");
    OsaConsoleExit();
    OtxDrvReleaseLib();
    exit(1);
    break;
  case OTX_RES_NOT_SUPPORTED:
    DemoLogf("Error: Mismatch between OTX Driver revision and OTX Library revision.\nRevisions must match.\n");
    OsaConsoleExit();
    OtxDrvReleaseLib();
    exit(1);
    break;
  default:
    DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
    break;
  }

  // Query driver revision
  {
    char szVersionInfo[255];
    result = OtxDrvGetAttributeValueSz(OTX_DRIVER_DEVICE_HANDLE, OTX_ATTR_DRV_REV_STR, szVersionInfo, 255); DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
    
    DemoLogf("Driver version %s\n", szVersionInfo);
  }

  if (g_demoEnv.nBoardType == OTX_DEVICE_UNKNOWN) {
    // Board type was not supplied by a command line parameter, so enumerate and find a board
    result = FindBoardDevice(&g_demoEnv.nBoardType);
    if (result == OTX_S_FALSE) {
      DemoLogf("No OTX board was found\n");
      DemoExitQuestion();
    } else if (result != OTX_S_OK) {
      DemoExitQuestion();
    }
  }

  // Connect (get handles) to the physical devices on the OTX board
  DemoLogf("Get Handle to the board device...\n");
  DemoOpenPhysicalDevices();

  DemoLogf("Read Serial Number to the board device...\n");


  // Read Serial Number from the board
  {
    OtxAttribTypeE  eAttribType;    // Type of the attribute, if attribute exists
    OTX_UINT32      cbAttribValue;  // Size in bytes of attribute value (Get value with <f OtxDrvGetAttributeValue>)
    OTX_BOOL        bReadOnly;      // Is attribute writable or read-only
    char *pszSerialNumber;

    result = OtxDrvGetAttributeType(g_hPhysDevs.hThorController0, OTX_ATTR_BRD_SERIAL_NUMBER, &eAttribType, &cbAttribValue, &bReadOnly); DemoCheckRc(g_hPhysDevs.hThorController0, result);
    assert(eAttribType == OTX_ATTRIB_SZ);
    assert(bReadOnly == OTX_TRUE);
    pszSerialNumber = (char *) malloc(cbAttribValue * sizeof(char));
    result = OtxDrvGetAttributeValueSz(g_hPhysDevs.hThorController0, OTX_ATTR_BRD_SERIAL_NUMBER, pszSerialNumber, cbAttribValue); DemoCheckRc(g_hPhysDevs.hThorController0, result);
    DemoLogf("Board Serial Number: %s\n", pszSerialNumber);
    free(pszSerialNumber);
  }

  // Read Build Number from the board
  {
    OtxAttribTypeE  eAttribType;    // Type of the attribute, if attribute exists
    OTX_UINT32      cbAttribValue;  // Size in bytes of attribute value (Get value with <f OtxDrvGetAttributeValue>)
    OTX_BOOL        bReadOnly;      // Is attribute writable or read-only
    char *pszBrdBuildNumber;

    result = OtxDrvGetAttributeType(g_hPhysDevs.hThorController0, OTX_ATTR_BRD_BUILD, &eAttribType, &cbAttribValue, &bReadOnly); DemoCheckRc(g_hPhysDevs.hThorController0, result);
    assert(eAttribType == OTX_ATTRIB_SZ);
    assert(bReadOnly == OTX_TRUE);
    pszBrdBuildNumber = (char *) malloc(cbAttribValue * sizeof(char));
    result = OtxDrvGetAttributeValueSz(g_hPhysDevs.hThorController0, OTX_ATTR_BRD_BUILD, pszBrdBuildNumber, cbAttribValue); DemoCheckRc(g_hPhysDevs.hThorController0, result);
    DemoLogf("Board Build Number: %s\n", pszBrdBuildNumber);
    free(pszBrdBuildNumber);
  }

  if (g_demoEnv.nBoardType != OTX_DEVICE_THOR_PCMCIA) {
    // Read CPLD Build Numbers from the board
    OtxAttribTypeE  eAttribType;    // Type of the attribute, if attribute exists
    OTX_UINT32      cbAttribValue;  // Size in bytes of attribute value (Get value with <f OtxDrvGetAttributeValue>)
    OTX_BOOL        bReadOnly;      // Is attribute writable or read-only
    char *pszCpldBuildNumbers;
    
    result = OtxDrvGetAttributeType(g_hPhysDevs.hThorController0, OTX_ATTR_BRD_CPLD_BUILD, &eAttribType, &cbAttribValue, &bReadOnly); DemoCheckRc(g_hPhysDevs.hThorController0, result);
    assert(eAttribType == OTX_ATTRIB_SZ);
    assert(bReadOnly == OTX_TRUE);
    pszCpldBuildNumbers = (char *) malloc(cbAttribValue * sizeof(char));
    result = OtxDrvGetAttributeValueSz(g_hPhysDevs.hThorController0, OTX_ATTR_BRD_CPLD_BUILD, pszCpldBuildNumbers, cbAttribValue); DemoCheckRc(g_hPhysDevs.hThorController0, result);
    DemoLogf("CPLD Build:\n%s", pszCpldBuildNumbers);
    free(pszCpldBuildNumbers);
  }

  // Main Loop. Loop here and wait for Events from the driver or commands from the user ....

  DemoLogf("\nPress 'h' for help\n");

  while (!exitFlag) {
    commandChar = OsaConsoleGetCh();
    exitFlag = DemoProcessUserCommand(commandChar);
  }

  // Clean up
  OtxDrvReleaseLib();
  OsaConsoleExit();

  return 0;
}

